package fr.lri.hcc.ilda.touchtokens;

import android.util.Log;

import java.text.DecimalFormat;
import java.util.ArrayList;

/*
 *   Authors: Caroline Appert (caroline.appert@lri.fr)
 *   Copyright (c) Univ. Paris-Sud XI, 2016. All Rights Reserved
 *   Licensed under the GNU LGPL. For full terms see the file COPYING.
*/

public abstract class AbstractBuffer {

    protected ArrayList<Double> distance;
    protected ArrayList<Double> time;
    protected ArrayList<Integer> fingerCount;

    protected int window = 80;

    public AbstractBuffer() {
        distance = new ArrayList<Double>();
        time = new ArrayList<Double>();
        fingerCount = new ArrayList<Integer>();
    }

    public synchronized void updateBuffer(double now, double value, int fc) {
        ArrayList<Double> clearDist = new ArrayList<Double>();
        ArrayList<Double> clearTime = new ArrayList<Double>();
        ArrayList<Integer> clearFinger = new ArrayList<Integer>();
        for (int i = 0; i < distance.size(); i++) {
            if( (now - this.time.get(i)) <= window) {
                clearDist.add(this.distance.get(i));
                clearTime.add(this.time.get(i));
                clearFinger.add(this.fingerCount.get(i));
            }
        }
        clearDist.add(value);
        clearTime.add(now);
        clearFinger.add(fc);
        this.time = clearTime;
        this.distance = clearDist;
        this.fingerCount = clearFinger;
    }

    public static DecimalFormat DEC_3_FORMATTER = new DecimalFormat("##.###");

    public void printBufferStats() {
        Log.v("TOUCHTOKENS", "buffer stats: ["+min(distance)+"-"+max(distance)+"] ~"+mean(distance)+" / "+distance.size()+" points");
        String str = "";
        for(int i = 0; i < distance.size(); i++) {
            str += DEC_3_FORMATTER.format(distance.get(i))+" ["+fingerCount.get(i)+"], ";
        }
        Log.v("TOUCHTOKENS", str);
    }

    private double max(ArrayList<Double> bufferValue) {
        double max = Double.MIN_VALUE;
        for (int i = 0; i < bufferValue.size(); i++) {
            max = Math.max(bufferValue.get(i), max);
        }
        return max;
    }

    private double min(ArrayList<Double> bufferValue) {
        double min = Double.MAX_VALUE;
        for (int i = 0; i < bufferValue.size(); i++) {
            min = Math.min(bufferValue.get(i), min);
        }
        return min;
    }

    private double mean(ArrayList<Double> bufferValue) {
        double sum = 0;
        for (int i = 0; i < bufferValue.size(); i++) {
            sum += bufferValue.get(i);
        }
        return sum / bufferValue.size();
    }

    protected double getSlope(int fc) {
        TheilRegression sr = new TheilRegression();
        ArrayList<Double> X = time;
        ArrayList<Double> Y = distance;
        ArrayList<Double> restY = new ArrayList<Double>();
        ArrayList<Double> restX = new ArrayList<Double>();
        for (int i = 0; i < Y.size(); i++) {
            if (fingerCount.get(i) == fc) {
                restY.add(Y.get(i));
                restX.add(X.get(i));
            }
        }
        if(restX.size() >= 2) {
            try {
                double[] result = sr.getTheilRegression(restX, restY);
                return result[1];
            } catch(Exception e) {
                return 0;
            }
        } else {
            return 0;
        }
    }

    public void bufferClear() {
        distance.clear();
        time.clear();
        fingerCount.clear();
    }
}
