package fr.lri.hcc.ilda.touchtokens;

/*
 *   Authors: Caroline Appert (caroline.appert@lri.fr), Rafael Morales (Rafael.Morales@inria.fr)
 *   Copyright (c) Univ. Paris-Sud XI, 2016. All Rights Reserved
 *   Licensed under the GNU LGPL. For full terms see the file COPYING.
 */

import android.util.Log;

import java.util.ArrayList;
import java.util.Iterator;

public class BufferSqueeze {

	private ArrayList<Double> distance ;
	private ArrayList<Double> time ;

	private double dRef = 0;
	private double threshold ;

	protected int window = 40;
	
	private boolean squeezed = false;

	private SqueezeState state;

	private int holdingFingerCount;

	private boolean toInit = true;

	public BufferSqueeze() {
		distance = new ArrayList<Double>();
		time = new ArrayList<Double>();
		// based on or data this is the threshold to detect a squeeze
//		threshold = 0.85;
		threshold = 1.05;

	}

	public ArrayList<Double> getBuffer() {
		return distance;
	}

	public synchronized void updateBuffer(double now, double value) {
		ArrayList<Double> clearDist = new ArrayList<Double>();
		ArrayList<Double> clearTime = new ArrayList<Double>();
		for (int i = 0; i < distance.size(); i++) {
			if((now - this.time.get(i)) <= window) {
				clearDist.add(this.distance.get(i));
				clearTime.add(this.time.get(i));
			}
		}
		clearDist.add(value);
		clearTime.add(now);
		this.time = clearTime;
		this.distance = clearDist;

		boolean wasSqueezed = squeezed;
		squeezed = ((dRef - max(distance)) > threshold);
		if(squeezed && !wasSqueezed) {
			state = SqueezeState.INITSQUEEZE;
		} else if(!squeezed && wasSqueezed) {
			state = SqueezeState.ENDSQUEEZE;
		} else if(squeezed) {
			state = SqueezeState.SQUEEZING;
		} else {
			state = SqueezeState.NONE;
		}
//		printBufferStats();
	}

	public void printBufferStats() {
		Log.v("TOUCHTOKENS", "buffer stats: ["+min(distance)+"-"+max(distance)+"] ~"+mean(distance)+" / "+distance.size()+" points");
		String str = "";
		for (Iterator<Double> iterator = distance.iterator(); iterator.hasNext();) {
			Double y = (Double) iterator.next();
			str += y+", ";
		}
		Log.v("TOUCHTOKENS", str);
	}

	public ArrayList<Double> getDist() {
		return distance;
	}

	public ArrayList<Double> getTime() {
		return time;
	}

	public void bufferClear() {
		// cleaning the buffers...
		distance.clear();
		time.clear();
	}

	private double max(ArrayList<Double> bufferValue) {
		double max = Double.MIN_VALUE;
		for (int i = 0; i < bufferValue.size(); i++) {
			max = Math.max(bufferValue.get(i), max);
		}
		return max;
	}

	private double min(ArrayList<Double> bufferValue) {
		double min = Double.MAX_VALUE;
		for (int i = 0; i < bufferValue.size(); i++) {
			min = Math.min(bufferValue.get(i), min);
		}
		return min;
	}

	private double mean(ArrayList<Double> bufferValue) {
		double sum = 0;
		for (int i = 0; i < bufferValue.size(); i++) {
			sum += bufferValue.get(i);
		}
		return sum / bufferValue.size();
	}

	public void init(double dRef, int holdingFingerCount) {
		this.dRef = dRef;
		this.holdingFingerCount = holdingFingerCount;
		toInit = false;
		bufferClear();
	}

	public boolean isSqueezed() {
		return squeezed;
	}

	public void setSqueezed(boolean squeezed) {
		this.squeezed = squeezed;
	}

	public SqueezeState getState() {
		return state;
	}

	public void setState(SqueezeState state) {
		this.state = state;
	}

	public int getHoldingFingerCount() {
		return holdingFingerCount;
	}

	public void setHoldingFingerCount(int holdingFingerCount) {
		this.holdingFingerCount = holdingFingerCount;
	}

	public boolean isToInit() {
		return toInit;
	}

	public void setToInit(boolean toInit) {
		this.toInit = toInit;
	}

}
