package fr.lri.hcc.ilda.touchtokens;

/*
 *   Authors: Caroline Appert (caroline.appert@lri.fr)
 *   Copyright (c) Univ. Paris-Sud XI, 2016. All Rights Reserved
 *   Licensed under the GNU LGPL. For full terms see the file COPYING.
*/

import android.app.Activity;
import android.content.Context;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Paint;
import android.graphics.Rect;
import android.util.AttributeSet;
import android.util.Log;
import android.view.MotionEvent;

import java.text.DecimalFormat;
import java.util.ArrayList;

/**
 * Created by appert on 03/09/15.
 */
public class TestView extends TouchTokenView {

    protected String tokenID;
    protected ArrayList<TouchPoint> touchPoints;
    protected ArrayList<TouchPoint> notchPoints;
    protected TokenTemplate transformedTemplate;
    protected boolean isGeometryUncertain;

    protected boolean bend = false;
    protected boolean squeezed = false;

    protected String eventID = "";

    protected Paint recoPaint = new Paint();
    protected Paint touchPaint = new Paint();
    protected Paint tokenPaint = new Paint();
    protected Paint tokenCenterPaint = new Paint();
    protected Paint tokenOutlinePaint = new Paint();
    protected Paint tokenPaintBend = new Paint();
    protected Paint tokenPaintSqueezed = new Paint();

    protected DecimalFormat df = new DecimalFormat("0.000");

    public TestView(Context context) {
        super(context);
        initPaints();
    }

    public TestView(Context context, AttributeSet attrs) {
        super(context, attrs);
        initPaints();
    }

    protected void initPaints() {
        touchPaint.setColor(Color.LTGRAY);
        touchPaint.setStyle(Paint.Style.FILL);
        touchPaint.setAntiAlias(true);
        tokenPaint.setColor(Color.DKGRAY);
        tokenPaint.setStyle(Paint.Style.STROKE);
        tokenPaint.setAntiAlias(true);
        tokenCenterPaint.setColor(Color.BLUE);
        tokenCenterPaint.setStyle(Paint.Style.FILL);
        tokenCenterPaint.setAntiAlias(true);
        tokenOutlinePaint.setColor(Color.BLUE);
        tokenOutlinePaint.setStyle(Paint.Style.STROKE);
        tokenOutlinePaint.setStrokeWidth(0); // hairline mode
        tokenOutlinePaint.setAntiAlias(true);
        tokenPaintBend.setColor(Color.BLUE);
        tokenPaintBend.setStyle(Paint.Style.FILL);
        tokenPaintBend.setAlpha(50);
        tokenPaintBend.setAntiAlias(true);
        tokenPaintSqueezed.setColor(Color.BLUE);
        tokenPaintSqueezed.setStyle(Paint.Style.FILL);
        tokenPaintSqueezed.setAntiAlias(true);
        recoPaint.setColor(Color.BLACK);
        recoPaint.setTextSize(40);
    }

    private void updateInfoFromLastEvent(String tokenID, ArrayList<TouchPoint> touchPoints, ArrayList<TouchPoint> notchPoints, double viewWidthMm, double viewHeightMm, TokenTemplate transformedTemplate, boolean insufficientInfo) {
        this.tokenID = tokenID;
        this.touchPoints = touchPoints;
        this.notchPoints = notchPoints;
        this.transformedTemplate = transformedTemplate;
        this.isGeometryUncertain = insufficientInfo;
        Activity host = (Activity) this.getContext();
        host.runOnUiThread(new Runnable() {
            @Override
            public void run() {
                invalidate();
            }
        });
    }

    @Override
    public void onTokenDown(String tokenID, ArrayList<TouchPoint> touchPoints, ArrayList<TouchPoint> notchPoints, double viewWidthMm, double viewHeightMm, TokenTemplate transformedTemplate, boolean insufficientInfo) {
        updateInfoFromLastEvent(tokenID, touchPoints, notchPoints, viewWidthMm, viewHeightMm, transformedTemplate, insufficientInfo);
    }

    @Override
    public void onTokenMoved(String tokenID, ArrayList<TouchPoint> touchPoints, ArrayList<TouchPoint> notchPoints, double viewWidthMm, double viewHeightMm, TokenTemplate transformedTemplate, boolean insufficientInfo) {
        updateInfoFromLastEvent(tokenID, touchPoints, notchPoints, viewWidthMm, viewHeightMm, transformedTemplate, insufficientInfo);
    }

    @Override
    public void onTokenUp(String tokenID, ArrayList<TouchPoint> touchPoints, ArrayList<TouchPoint> notchPoints, double viewWidthMm, double viewHeightMm, TokenTemplate transformedTemplate, boolean insufficientInfo) {
        updateInfoFromLastEvent(null, touchPoints, notchPoints, viewWidthMm, viewHeightMm, transformedTemplate, insufficientInfo);
    }

    @Override
    public void onTokenBent(String tokenID, ArrayList<TouchPoint> touchPoints, ArrayList<TouchPoint> notchPoints, double viewWidthMm, double viewHeightMm, TokenTemplate transformedTemplate, boolean insufficientInfo) {
        this.bend = true;
        Log.v("debugging", "onTokenBent " + bend);
        updateInfoFromLastEvent(tokenID, touchPoints, notchPoints, viewWidthMm, viewHeightMm, transformedTemplate, insufficientInfo);
    }

    @Override
    public void onTokenUnbent(String tokenID, ArrayList<TouchPoint> touchPoints, ArrayList<TouchPoint> notchPoints, double viewWidthMm, double viewHeightMm, TokenTemplate transformedTemplate, boolean insufficientInfo) {
        this.bend = false;
        Log.v("debugging", "onTokenUnbent " + bend);
        updateInfoFromLastEvent(tokenID, touchPoints, notchPoints, viewWidthMm, viewHeightMm, transformedTemplate, insufficientInfo);
    }

    @Override
    public void onTokenPressed(String tokenID, ArrayList<TouchPoint> touchPoints, ArrayList<TouchPoint> notchPoints, double viewWidthMm, double viewHeightMm, TokenTemplate transformedTemplate, boolean insufficientInfo) {
        this.squeezed = true;
        updateInfoFromLastEvent(tokenID, touchPoints, notchPoints, viewWidthMm, viewHeightMm, transformedTemplate, insufficientInfo);
    }

    @Override
    public void onTokenReleased(String tokenID, ArrayList<TouchPoint> touchPoints, ArrayList<TouchPoint> notchPoints, double viewWidthMm, double viewHeightMm, TokenTemplate transformedTemplate, boolean insufficientInfo) {
        this.squeezed = false;
        updateInfoFromLastEvent(tokenID, touchPoints, notchPoints, viewWidthMm, viewHeightMm, transformedTemplate, insufficientInfo);
    }

    public void fingerUp(MotionEvent event, int index, int id) {
        super.fingerUp(event, index, id);

        int fingerCount = 0;
        for (int i = 0; i < FINGERS_LOCATION.length; i++) {
            if(FINGERS_LOCATION[i] != null) {
                fingerCount++;
            }
        }
        if(fingerCount == 0 && touchPoints != null) {
            touchPoints.clear();
        }

        Activity host = (Activity) this.getContext();
        host.runOnUiThread(new Runnable() {
            @Override
            public void run() {
                invalidate();
            }
        });
    }

    @Override
    protected void onDraw(Canvas canvas) {
        super.onDraw(canvas);

        if (tokenID == null && touchPoints != null && transformedTemplate != null) {
            return;
        }
        synchronized (this) {
            if (tokenID == null) {
                return;
            }

            eventID = "";

            viewWidthMm = getWidth() / MainActivity.SCREEN_RESOLUTION;
            viewHeightMm = getHeight() / MainActivity.SCREEN_RESOLUTION;

            TouchPoint tokenCenter = null;
            double tokenRelOrientation = 0;
            double tokenAbsOrientation = 0;

            if (touchPoints != null) {
                for (int i = 0; i < touchPoints.size(); i++) {
                    if (touchPoints.get(i) != null) {
                        canvas.drawCircle(
                                (int) (touchPoints.get(i).getX() * getWidth()),
                                (int) (touchPoints.get(i).getY() * getHeight()),
                                40, touchPaint);
                    }
                }
            }

            if (notchPoints != null) {
                for (int i = 0; i < notchPoints.size(); i++) {
                    if (notchPoints.get(i) != null) {
                        canvas.drawCircle(
                                (int) (notchPoints.get(i).getX() * getWidth()),
                                (int) (notchPoints.get(i).getY() * getHeight()),
                                40, tokenPaint);
                    }
                }
            }

            if (transformedTemplate != null) {
                tokenCenter = getTokenCenter(transformedTemplate, notchPoints);
                tokenRelOrientation = getTokenRelOrientation(transformedTemplate, notchPoints);
                tokenAbsOrientation = getTokenInitOrientation(transformedTemplate) + tokenRelOrientation;
                canvas.save();
                canvas.drawCircle(
                        (int) (tokenCenter.x * getWidth()),
                        (int) (tokenCenter.y * getHeight()),
                        40, tokenCenterPaint);
                canvas.translate((float) (tokenCenter.x * getWidth()), (float) (tokenCenter.y * getHeight()));
                canvas.rotate(-(float) (tokenAbsOrientation * 180 / Math.PI));

                if(bend) {
                    eventID = "bending";
                    canvas.drawPath(transformedTemplate.getOriginalGeometryTemplate().getContourShape(), tokenPaintBend);
                } else {
                    if(squeezed) {
                        eventID = "squeezing";
                        canvas.drawPath(transformedTemplate.getOriginalGeometryTemplate().getContourShape(), tokenPaintSqueezed);
                    } else {
                        canvas.drawPath(transformedTemplate.getOriginalGeometryTemplate().getContourShape(), tokenOutlinePaint);
                    }
                }

                canvas.restore();
            }

            String textTokenID = tokenID;
            String textLocation = "(" + (int) (tokenCenter.x * getWidth()) + ", " + (int) (tokenCenter.y * getHeight()) + ")";
            String textOrientation = "abs=" + df.format(tokenAbsOrientation) + ", rel=" + df.format(tokenRelOrientation);
            Rect boundsTextTokenID = new Rect();
            recoPaint.getTextBounds(textTokenID, 0, textTokenID.length(), boundsTextTokenID);
            Rect boundsTextLocation = new Rect();
            recoPaint.getTextBounds(textLocation, 0, textLocation.length(), boundsTextLocation);
            Rect boundsTextOrientation = new Rect();
            recoPaint.getTextBounds(textOrientation, 0, textOrientation.length(), boundsTextOrientation);
            canvas.drawText(textTokenID, 10, getHeight() - 10 - boundsTextOrientation.height() - 10 - boundsTextLocation.height() - 10 - boundsTextTokenID.height(), recoPaint);
            canvas.drawText(textLocation, 10, getHeight() - 10 - boundsTextOrientation.height() - 10 - boundsTextLocation.height(), recoPaint);
            canvas.drawText(textOrientation, 10, getHeight() - 10 - boundsTextOrientation.height(), recoPaint);
        }
    }
}
