package fr.lri.hcc.ilda.touchtokens;
/*
 *   Source: project molgenis https://github.com/molgenis 
 *   https://github.com/molgenis/systemsgenetics/blob/master/genetica-libraries/src/main/java/umcg/genetica/math/stats/TheilSen.java
 *
 */

import android.util.Log;

import java.util.ArrayList;

public class TheilRegression {

	public TheilRegression() { }


	/**
	 * Used by the quick sort and quick select algorithms.
	 */
	private static void swap(final double a[], final int i, final int j)
	{
		final double T;
		T = a[i];
		a[i] = a[j];
		a[j] = T;

	}

	/**
	 * Returns a copy of the array.
	 */
	public static double[] copy(final double[] v) {
		double[] ans = new double[v.length];
		System.arraycopy(v,0,ans,0,v.length);
		return(ans);
	}
	/**
	 * Returns a copy of the array.
	 */
	public static double[][] copy(double[][] v) {
		double[][] ans = new double[v.length][];
		for (int k=0;k<v.length;k++)
			ans[k]=copy(v[k]);
		return(ans);
	}

	/*
        * http://ndevilla.free.fr/median/median/index.html
        * quickselect.c
        * @author N. Devillard
        */
	private static double quickSelect(double[] v, boolean upperMedian) {
		int low=0, high=v.length-1;
		final int median = upperMedian ? high/2+1 : high/2;
		int middle, ll, hh;

		for (;;) {
			if (high <= low) { /* One element only */
				return v[median];
			}

			if (high == low + 1) {  /* Two elements only */
				if (v[low] > v[high])
					swap(v, low, high);
				return v[median];
			}

                /* Find median of low, middle and high items; swap into position low */
			middle = (low + high) / 2;
			if (v[middle] > v[high]) swap(v, middle, high);
			if (v[low] > v[high]) swap(v, low, high);
			if (v[middle] > v[low]) swap(v, middle, low);

                /* Swap low item (now in position middle) into position (low+1) */
			swap(v, middle, low+1);

                /* Nibble from each end towards middle, swapping items when stuck */
			ll = low + 1;
			hh = high;
			for (;;) {
				do ll++; while (v[low] > v[ll]);
				do hh--; while (v[hh]  > v[low]);

				if (hh < ll)
					break;

				swap(v, ll, hh);
			}

                /* Swap middle item (in position low) back into correct position */
			swap(v, low, hh);

                /* Re-set active partition */
			if (hh <= median)
				low = ll;
			if (hh >= median)
				high = hh - 1;
		}
	}

	/*
* http://ndevilla.free.fr/median/median/index.html
* optmed.c
* @author N. Devillard
*/
	private static void sortInPlace(final double[] v, final int i, final int j) {
		if(v[i] > v[j]) {
			final double tmp = v[i];
			v[i] = v[j];
			v[j] = tmp;
		}
	}

	private static double median3(double[] v) {
		sortInPlace(v,0,1);
		sortInPlace(v,1,2);
		sortInPlace(v,0,1);
		return v[1];
	}
	private static double median5(double[] v) {
		sortInPlace(v,0,1);
		sortInPlace(v,3,4);
		sortInPlace(v,0,3);
		sortInPlace(v,1,4);
		sortInPlace(v,1,2);
		sortInPlace(v,2,3);
		sortInPlace(v,1,2);
		return v[2];
	}
	private static double median7(double[] v) {
		sortInPlace(v,0,5);
		sortInPlace(v,0,3);
		sortInPlace(v,1,6);
		sortInPlace(v,2,4);
		sortInPlace(v,0,1);
		sortInPlace(v,3,5);
		sortInPlace(v,2,6);
		sortInPlace(v,2,3);
		sortInPlace(v,3,6);
		sortInPlace(v,4,5);
		sortInPlace(v,1,4);
		sortInPlace(v,1,3);
		sortInPlace(v,3,4);
		return v[3];
	}

	/**
	 * Computes the median of an array.
	 * If the array contains an even number of elements,
	 * then the mean of the lower and upper medians is returned.
	 */
	public static double median(double[] v) {
		if(v.length == 3) {
			return median3(copy(v));
		} else if(v.length == 5) {
			return median5(copy(v));
		} else if(v.length == 7) {
			return median7(copy(v));
		} else if(v.length % 2 == 1) { // odd
			return quickSelect(copy(v), false);
		} else { // even
			double[] tmp = copy(v);
			double lowerMedian = quickSelect(tmp, false);
			double upperMedian = quickSelect(tmp, true);
			return (lowerMedian+upperMedian)/2.0;
		}
	}

	/**
	 * This is a generic version of C.A.R Hoare's Quick Sort
	 * algorithm.  This will handle arrays that are already
	 * sorted, and arrays with duplicate keys.<BR>
	 *
	 * If you think of a one dimensional array as going from
	 * the lowest index on the left to the highest index on the right
	 * then the parameters to this function are lowest index or
	 * left and highest index or right.  The first time you call
	 * this function it will be with the parameters 0, a.length - 1.
	 * (taken out of a code by James Gosling and Kevin A. Smith provided
	 * with Sun's JDK 1.1.7)
	 *
	 * @param a       a double array
	 * @param lo0     left boundary of array partition (inclusive).
	 * @param hi0     right boundary of array partition (inclusive).
	 * @deprecated
	 */
	private static void quickSortMinToMax(double a[], int lo0, int hi0)
	{
		int lo = lo0;
		int hi = hi0;
		double mid;

		if ( hi0 > lo0)
		{

         /* Arbitrarily establishing partition element as the midpoint of
          * the array.
          */
			mid = a[(int) Math.round( ( lo0 + hi0 ) / 2.0) ];

			// loop through the array until indices cross
			while( lo <= hi )
			{
            /* find the first element that is greater than or equal to
             * the partition element starting from the left Index.
             */
				while( ( lo < hi0 )  && ( a[lo] < mid ))
					++lo;

            /* find an element that is smaller than or equal to
             * the partition element starting from the right Index.
             */
				while( ( hi > lo0 )  && ( a[hi] > mid ))
					--hi;

				// if the indexes have not crossed, swap
				if( lo <= hi )
				{
					swap(a, lo, hi);
					++lo;
					--hi;
				}
			}

         /* If the right index has not reached the left side of array
          * must now sort the left partition.
          */
			if( lo0 < hi )
				quickSortMinToMax( a, lo0, hi );

         /* If the left index has not reached the right side of array
          * must now sort the right partition.
          */
			if( lo < hi0 )
				quickSortMinToMax( a, lo, hi0 );

		}
	}

	/**
	 * Returns a sorted array from the minimum to the maximum value.
	 */
	public static double[] sortMinToMax(double[] v) {
		double[] ans=copy(v);
		quickSortMinToMax(ans, 0, ans.length - 1);
		return(ans);
	}

	/**
	 * Gives the percentile of an array.
	 * @param v an unsorted array
	 * @param p percentile, must be between 0 and 1
	 * @exception IllegalArgumentException if parameter.
	 * p is not between 0 and 1.
	 */
	public static double percentile(double[] v, double p) {
		return percentile(v, p, false);
	}
	public static double percentile(double[] v, double p, boolean isSorted) {
		if((p<0) || (p>1)) {
			throw new IllegalArgumentException("Percentile must be between 0 and 1 : "+p);
		}
		final double[] ans = (isSorted ? v : sortMinToMax(v));
		final int pos = (int) Math.floor(p*(ans.length-1));
		final double dif = p*(ans.length-1) - Math.floor(p*(ans.length-1));
		if(pos == (ans.length-1))
			return(ans[ans.length-1]);
		else
			return(ans[pos]*(1.0-dif) + ans[pos+1]*dif);
	}

	//     Calculates Theil-Sen estimation for given arrays.
	//     @return [y intercept, slope, 2.5 percentile, 97.5 percentile, number of pairs]
	public double[] getTheilRegression(ArrayList<Double> v1, ArrayList<Double> v2) {

        Log.v("TOUCHTOKENS", "v1.size()="+v1.size());

		if (v1.size() != v2.size()) {
			throw new IllegalArgumentException("Arrays must be of the same length " + v1.size() + ", " + v2.size());
		}

		ArrayList<Double> slopesList = new ArrayList<Double>();
		int cnt = 0;
		for (int i = 0; i < v1.size(); i++) {
			double x = v1.get(i);
			double y = v2.get(i);
			for (int j = i + 1; j < v1.size(); j++) {
				if (x != v1.get(j)) { // x must be different, otherwise slope becomes infinite
					double slope = (v2.get(j) - y) / (v1.get(j) - x);
					slopesList.add(slope);
					++cnt;
				}
			}
		}


        if (cnt == 0) {
            throw new IllegalArgumentException("cannot estimate slope");
        }

		Log.v("TOUCHTOKENS", "cnt="+cnt);

		double[] slopes = new double[slopesList.size()];
		for(int i=0; i<slopesList.size(); i++)
		{
			slopes[i] = slopesList.get(i);
		}
		double[] medianv1 = new double[v1.size()];
		for(int i=0; i<v1.size(); i++) {
			medianv1[i] = v1.get(i);
		}
		double median1 = median(medianv1);
		double[] medianv2 = new double[v2.size()];
		for(int i=0; i<v2.size(); i++) {
			medianv2[i] = v2.get(i);
		}
		double median2 = median(medianv2);
		double slope = median(slopes);
		double yI = median2 - slope * median1;
		double p1 = percentile(slopes, 0.025d);
		double p2 = percentile(slopes, 0.975d);

		return new double[]{yI, slope, p1, p2, cnt};

	}

}
