package fr.lri.hcc.ilda.touchtokens;

    /*
     *   Authors: Caroline Appert (caroline.appert@lri.fr)
     *   Copyright (c) Univ. Paris-Sud XI, 2016. All Rights Reserved
     *   Licensed under the GNU LGPL. For full terms see the file COPYING.
    */

import android.content.Context;
import android.os.CountDownTimer;
import android.support.v4.view.MotionEventCompat;
import android.util.AttributeSet;
import android.util.Log;
import android.view.MotionEvent;
import android.view.View;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Timer;
import java.util.TimerTask;

public abstract class TouchTokenView extends View {

    protected String tokenDown = null;

    protected static TouchPoint[] FINGERS_LOCATION = new TouchPoint[10];
    protected static TouchPoint[] FINGERS_LOCATION_AT_TIMER_START_TIME = new TouchPoint[10];

    protected int dwellDuration = 200; // in ms
    protected double dwellTolerance = 3; // in mm

    // (R) I add a CountDownTimer for bending, we need it for discriminating bfore we remove the token on the surface
    protected CountDownTimer timerBending;
    protected boolean allowingStartTimer = false;

    protected Timer dwellTimer;
    protected long dwellStartTime = -1;
    protected long lastUpdateTime = -1;

    protected boolean debug = false;

    protected double viewWidthMm = -1;
    protected double viewHeightMm = -1;

    protected ArrayList<TouchPoint> notchPointsAtDown;
    protected ArrayList<TouchPoint> currentPoint;
    protected TokenTemplate currentTemplate;


    protected BufferBend bufferBend = new BufferBend();
    protected BufferSqueeze bufferSqueeze = new BufferSqueeze();
    protected BufferSurface bufferSurface = new BufferSurface();


    protected boolean insufficientInfo = false;
    protected boolean bending = false;
    protected long timeLastEvent = -1;
    protected ArrayList<TouchPoint> lastReliableTokenPoints = new ArrayList<TouchPoint>();
    protected int lastPointCount = 0;

    protected Timer mtEventFeeder;

    public TouchTokenView(Context context) {
        super(context);
        init();
    }

    public TouchTokenView(Context context, AttributeSet attrs) {
        super(context, attrs);
        init();
    }

    public TouchTokenView(Context context, AttributeSet attrs, int defStyleAttr) {
        super(context, attrs, defStyleAttr);
        init();
    }


    protected void init() {

        viewWidthMm = getWidth() / MainActivity.SCREEN_RESOLUTION;
        viewHeightMm = getHeight() / MainActivity.SCREEN_RESOLUTION;

        dwellTimer = new Timer();
        dwellTimer.schedule(new TimerTask() {
            public void run() {
                if (dwellStartTime == -1) {
                    return;
                }
                if (tokenDown != null) {
                    return;
                }
                long currentTime = System.currentTimeMillis();
                if ((currentTime - dwellStartTime) >= dwellDuration && (lastUpdateTime - dwellStartTime) < dwellDuration) {
                    ArrayList<TouchPoint> inputInMms = new ArrayList<TouchPoint>(); // the recognizer stores templates in mm
                    ArrayList<TouchPoint> inputIn01 = new ArrayList<TouchPoint>();

                    if(viewWidthMm == 0 || viewHeightMm == 0) {
                        viewWidthMm = getWidth() / MainActivity.SCREEN_RESOLUTION;
                        viewHeightMm = getHeight() / MainActivity.SCREEN_RESOLUTION;
                    }

                    for (int i = 0; i < FINGERS_LOCATION.length; i++) {
                        if (FINGERS_LOCATION[i] != null) {
                            inputInMms.add(new TouchPoint(FINGERS_LOCATION[i].x * viewWidthMm, FINGERS_LOCATION[i].y * viewHeightMm));
                            inputIn01.add(new TouchPoint(FINGERS_LOCATION[i].x, FINGERS_LOCATION[i].y));
                        }
                    }
                    if (debug) {
                        Log.v("TOUCHTOKENS", "input: " + inputInMms);
                    }


                    currentTemplate = MainActivity.RECOGNIZER.recognize(inputInMms);

                    if (debug) {
                        Log.v("TOUCHTOKENS", "recognized as: " + currentTemplate);
                    }
                    if (currentTemplate != null) {
                        tokenDown = currentTemplate.getTokenID();
                        notchPointsAtDown = new ArrayList<TouchPoint>();
                        for (TouchPoint pt : currentTemplate.getPoints()) {
                            notchPointsAtDown.add(new TouchPoint(pt.x / viewWidthMm, pt.y / viewHeightMm));
                        }
                        ArrayList<TouchPoint> tokenPoints = new ArrayList<TouchPoint>(inputIn01);
//                        (R) allowing the timer if the user bend the token because the token is recognize
                        allowingStartTimer = true;
                        onTokenDown(tokenDown, inputIn01, tokenPoints, viewWidthMm, viewHeightMm, currentTemplate, insufficientInfo);
                        // init buffers
                        TouchPoint tokenCentroidMm = TouchTokenRecognizer.centroid(inputInMms);
                        int fingerCount = inputInMms.size();
                        double distance = sumDistances(inputInMms, tokenCentroidMm) / fingerCount;
                        bufferBend.bufferClear();
                        bufferSqueeze.init(distance, fingerCount);
                        bufferSurface.bufferClear();
                    }
                }
                lastUpdateTime = System.currentTimeMillis();
            }
        }, 0, 25);

        mtEventFeeder = new Timer();
        mtEventFeeder.schedule(new TimerTask() {
            public void run() {
                if(tokenDown == null)
                    return;
                ArrayList<TouchPoint> inputIn01 = new ArrayList<TouchPoint>();
                for (int i = 0; i < 3; i++) {
                    if(FINGERS_LOCATION[i] != null) {
                        inputIn01.add(new TouchPoint(FINGERS_LOCATION[i].x, FINGERS_LOCATION[i].y));
                    } else {
                        inputIn01.add(null);
                    }
                }
                processTokenEvent(tokenDown, currentTemplate, inputIn01, viewWidthMm, viewHeightMm);
            }
        }, 0, 10);

        // (R) I init here the timer
        timerBending = new CountDownTimer(70,10) {
            @Override
            public void onTick(long l) {

            }

            @Override
            public void onFinish() {
                Log.v("debugging", "FINISH");

                bending = bufferBend.bend();
                if(bending) {

                    onTokenBent(
                            tokenDown,
                            currentPoint,
                            lastReliableTokenPoints,
                            viewWidthMm,
                            viewHeightMm,
                            currentTemplate,
                            insufficientInfo);
                    Log.v("CountDown", "Bending");

                }
            }

        };
    }

//    protected void init() {
//
//        viewWidthMm = getWidth() / MainActivity.SCREEN_RESOLUTION;
//        viewHeightMm = getHeight() / MainActivity.SCREEN_RESOLUTION;
//
//        dwellTimer = new Timer();
//        dwellTimer.schedule(new TimerTask() {
//            public void run() {
//                if (dwellStartTime == -1) {
//                    return;
//                }
//                if (tokenDown != null) {
//                    return;
//                }
//                long currentTime = System.currentTimeMillis();
//                if ((currentTime - dwellStartTime) >= dwellDuration && (lastUpdateTime - dwellStartTime) < dwellDuration) {
//                    ArrayList<TouchPoint> inputInMms = new ArrayList<TouchPoint>(); // the recognizer stores templates in mm
//                    ArrayList<TouchPoint> inputIn01 = new ArrayList<TouchPoint>();
//
//                    if(viewWidthMm == 0 || viewHeightMm == 0) {
//                        viewWidthMm = getWidth() / MainActivity.SCREEN_RESOLUTION;
//                        viewHeightMm = getHeight() / MainActivity.SCREEN_RESOLUTION;
//                    }
//
//                    for (int i = 0; i < FINGERS_LOCATION.length; i++) {
//                        if (FINGERS_LOCATION[i] != null) {
//                            inputInMms.add(new TouchPoint(FINGERS_LOCATION[i].x * viewWidthMm, FINGERS_LOCATION[i].y * viewHeightMm));
//                            inputIn01.add(new TouchPoint(FINGERS_LOCATION[i].x, FINGERS_LOCATION[i].y));
//                        }
//                    }
//                    if (debug) {
//                        Log.v("TOUCHTOKENS", "input: " + inputInMms);
//                    }
//
//
//                    currentTemplate = MainActivity.RECOGNIZER.recognize(inputInMms);
//
//                    if (debug) {
//                        Log.v("TOUCHTOKENS", "recognized as: " + currentTemplate);
//                    }
//                    if (currentTemplate != null) {
//                        tokenDown = currentTemplate.getTokenID();
//                        notchPointsAtDown = new ArrayList<TouchPoint>();
//                        for (TouchPoint pt : currentTemplate.getPoints()) {
//                            notchPointsAtDown.add(new TouchPoint(pt.x / viewWidthMm, pt.y / viewHeightMm));
//                        }
//                        ArrayList<TouchPoint> tokenPoints = new ArrayList<TouchPoint>(inputIn01);
//                        onTokenDown(tokenDown, inputIn01, tokenPoints, viewWidthMm, viewHeightMm, currentTemplate, insufficientInfo);
//                        // init buffers
//                        TouchPoint tokenCentroidMm = TouchTokenRecognizer.centroid(inputInMms);
//                        int fingerCount = inputInMms.size();
//                        double distance = sumDistances(inputInMms, tokenCentroidMm) / fingerCount;
//                        bufferBend.bufferClear();
//                        bufferSqueeze.init(distance, fingerCount);
//                        bufferSurface.bufferClear();
//                    }
//                }
//                lastUpdateTime = System.currentTimeMillis();
//            }
//        }, 0, 25);
//
//        mtEventFeeder = new Timer();
//        mtEventFeeder.schedule(new TimerTask() {
//            public void run() {
//                if(tokenDown == null)
//                    return;
//                ArrayList<TouchPoint> inputIn01 = new ArrayList<TouchPoint>();
//                for (int i = 0; i < 3; i++) {
//                    if(FINGERS_LOCATION[i] != null) {
//                        inputIn01.add(new TouchPoint(FINGERS_LOCATION[i].x, FINGERS_LOCATION[i].y));
//                    } else {
//                        inputIn01.add(null);
//                    }
//                }
//                processTokenEvent(tokenDown, currentTemplate, inputIn01, viewWidthMm, viewHeightMm);
//            }
//        }, 0, 10);
//
//    }

    protected void complementNotchPoints(ArrayList<TouchPoint> notchPointsAtDown, ArrayList<TouchPoint> touchPoints, double deviceWidth, double deviceHeight) {
        int missingIndex = -1;
        TouchPoint notchDown1 = null;
        TouchPoint notch1 = null;
        TouchPoint notchDown2 = null;
        TouchPoint notch2 = null;
        TouchPoint notchDown3 = null;
        for (int i = 0; i < touchPoints.size(); i++) {
            if(touchPoints.get(i) == null) {
                missingIndex = i;
                notchDown3 = notchPointsAtDown.get(i);
                notchDown3 = new TouchPoint(notchDown3.x*deviceWidth, notchDown3.y*deviceHeight);
            } else {
                if(notchDown1 == null) {
                    notchDown1 = notchPointsAtDown.get(i);
                    notchDown1 = new TouchPoint(notchDown1.x*deviceWidth, notchDown1.y*deviceHeight);
                    notch1 = touchPoints.get(i);
                    notch1 = new TouchPoint(notch1.x*deviceWidth, notch1.y*deviceHeight);
                } else {
                    notchDown2 = notchPointsAtDown.get(i);
                    notchDown2 = new TouchPoint(notchDown2.x*deviceWidth, notchDown2.y*deviceHeight);
                    notch2 = touchPoints.get(i);
                    notch2 = new TouchPoint(notch2.x*deviceWidth, notch2.y*deviceHeight);
                }
            }
        }
        if(missingIndex == -1) {
            return;
        } else {
            TouchPoint vNotchesDown13 = new TouchPoint(notchDown3.x - notchDown1.x, notchDown3.y - notchDown1.y);
            TouchPoint vNotchesDown12 = new TouchPoint(notchDown2.x - notchDown1.x, notchDown2.y - notchDown1.y);
            double angleNotchesDown213 = TouchTokenRecognizer.angleBetweenVectors(vNotchesDown12, vNotchesDown13);

            TouchPoint vNotches12 = new TouchPoint(notch2.x - notch1.x, notch2.y - notch1.y);
            double angleNotchesx12 = TouchTokenRecognizer.angleBetweenVectors(new TouchPoint(1, 0), vNotches12);

            double lengthNotchesDown13 =
                    Math.sqrt((notchDown3.x - notchDown1.x) * (notchDown3.x - notchDown1.x) +
                            (notchDown3.y - notchDown1.y) * (notchDown3.y - notchDown1.y));

            TouchPoint notch3 = new TouchPoint(lengthNotchesDown13, 0);
            notch3.setLocation(notch3.x + notch1.getX(), notch3.y + notch1.getY());
            double angle = angleNotchesx12 + angleNotchesDown213;
            TouchTokenRecognizer.rotateBy(notch3, notch1, angle, notch3);
            notch3.setLocation(notch3.x / deviceWidth, notch3.y / deviceHeight);
            touchPoints.set(missingIndex, notch3);
        }
    }

    public abstract void onTokenDown(String tokenID, ArrayList<TouchPoint> touchPoints, ArrayList<TouchPoint> notchPoints, double viewWidthMm, double viewHeightMm, TokenTemplate transformedTemplate, boolean insufficientInfo);

    public abstract void onTokenMoved(String tokenID, ArrayList<TouchPoint> touchPoints, ArrayList<TouchPoint> notchPoints, double viewWidthMm, double viewHeightMm, TokenTemplate transformedTemplate, boolean insufficientInfo);

    public abstract void onTokenUp(String tokenID, ArrayList<TouchPoint> touchPoints, ArrayList<TouchPoint> notchPoints, double viewWidthMm, double viewHeightMm, TokenTemplate transformedTemplate, boolean insufficientInfo);

    public abstract void onTokenBent(String tokenID, ArrayList<TouchPoint> touchPoints, ArrayList<TouchPoint> notchPoints, double viewWidthMm, double viewHeightMm, TokenTemplate transformedTemplate, boolean insufficientInfo);

    public abstract void onTokenUnbent(String tokenID, ArrayList<TouchPoint> touchPoints, ArrayList<TouchPoint> notchPoints, double viewWidthMm, double viewHeightMm, TokenTemplate transformedTemplate, boolean insufficientInfo);

    public abstract void onTokenPressed(String tokenID, ArrayList<TouchPoint> touchPoints, ArrayList<TouchPoint> notchPoints, double viewWidthMm, double viewHeightMm, TokenTemplate transformedTemplate, boolean insufficientInfo);

    public abstract void onTokenReleased(String tokenID, ArrayList<TouchPoint> touchPoints, ArrayList<TouchPoint> notchPoints, double viewWidthMm, double viewHeightMm, TokenTemplate transformedTemplate, boolean insufficientInfo);

    public void fingerDown(MotionEvent event, int index, int id) {
//        Log.v("TOUCHTOKENS", "index="+index+", id="+id);
        FINGERS_LOCATION[id] = new TouchPoint(
                event.getX(index)/getWidth(),
                event.getY(index)/getHeight());
        for (int i = 0; i < FINGERS_LOCATION.length; i++) {
            FINGERS_LOCATION_AT_TIMER_START_TIME[i] = FINGERS_LOCATION[i];
        }
        dwellStartTime = System.currentTimeMillis();
        bufferSqueeze.setToInit(true);
    }

    public void fingerMove(MotionEvent event, int index, int id) {
        FINGERS_LOCATION[id] = new TouchPoint(
                event.getX(index)/getWidth(),
                event.getY(index)/getHeight());

        // sometimes, MT events are not delivered in the right order, avoid null exception:
        if(FINGERS_LOCATION_AT_TIMER_START_TIME[id] == null) {
            FINGERS_LOCATION_AT_TIMER_START_TIME[id] = FINGERS_LOCATION[id];
        }

        TouchPoint pointNow = new TouchPoint(
                FINGERS_LOCATION[id].x * viewWidthMm,
                FINGERS_LOCATION[id].y * viewHeightMm);
        TouchPoint pointAtTimerStartTime = new TouchPoint(
                FINGERS_LOCATION_AT_TIMER_START_TIME[id].x * viewWidthMm,
                FINGERS_LOCATION_AT_TIMER_START_TIME[id].y * viewHeightMm);

        double distance = pointNow.distance(pointAtTimerStartTime);

        boolean dwelling = distance < dwellTolerance;
        if(!dwelling) {
            for (int i = 0; i < FINGERS_LOCATION.length; i++) {
                FINGERS_LOCATION_AT_TIMER_START_TIME[i] = FINGERS_LOCATION[i];
            }
            dwellStartTime = System.currentTimeMillis();
        }

        ArrayList<TouchPoint> inputIn01 = new ArrayList<TouchPoint>();
        for (int i = 0; i < FINGERS_LOCATION.length; i++) {
            if(FINGERS_LOCATION[i] != null) {
                inputIn01.add(new TouchPoint(FINGERS_LOCATION[i].x, FINGERS_LOCATION[i].y));
            } else {
                inputIn01.add(null);
            }
        }
        if(tokenDown != null) {
            ArrayList<TouchPoint> notchPoints = new ArrayList<TouchPoint>();
            for (int i = 0; i < 3; i++) {
                TouchPoint touchPoint = inputIn01.get(i);
                notchPoints.add(touchPoint);
            }
            processTokenEvent(tokenDown, currentTemplate, notchPoints, viewWidthMm, viewHeightMm);
        }
    }

//    protected void tokenUp() {
//        if(tokenDown != null) {
//            dwellStartTime = -1;
//            boolean stillOnSurface = false;
//            stillOnSurface = bufferSurface.onSurface();
//            if(!stillOnSurface) {
//                ArrayList<TouchPoint> inputIn01 = new ArrayList<TouchPoint>();
//                for (int i = 0; i < FINGERS_LOCATION.length; i++) {
//                    if(FINGERS_LOCATION[i] != null) {
//                        inputIn01.add(new TouchPoint(FINGERS_LOCATION[i].x, FINGERS_LOCATION[i].y));
//                    }
//                }
//                onTokenUp(tokenDown, inputIn01, inputIn01, viewWidthMm, viewHeightMm, currentTemplate, insufficientInfo);
//                tokenDown = null;
//            }
//
//        }
//    }


//    (R) when remove the token I cancel the timer
protected void tokenUp() {
    //(R) here the only line I added
    timerBending.cancel();
    if(tokenDown != null) {
        dwellStartTime = -1;
        boolean stillOnSurface = false;
        stillOnSurface = bufferSurface.onSurface();
        if(!stillOnSurface) {
            ArrayList<TouchPoint> inputIn01 = new ArrayList<TouchPoint>();
            for (int i = 0; i < FINGERS_LOCATION.length; i++) {
                if(FINGERS_LOCATION[i] != null) {
                    inputIn01.add(new TouchPoint(FINGERS_LOCATION[i].x, FINGERS_LOCATION[i].y));
                }
            }
            onTokenUp(tokenDown, inputIn01, inputIn01, viewWidthMm, viewHeightMm, currentTemplate, insufficientInfo);
            tokenDown = null;
        }

    }
}

//    public void fingerUp(MotionEvent event, int index, int id) {
//        int cursorID = id;
//        int fingerCount = 0;
//        for (int i = 0; i < FINGERS_LOCATION.length; i++) {
//            if(FINGERS_LOCATION[i] != null) {
//                fingerCount++;
//            }
//        }
////        Log.v("TOUCHTOKENS", "Finger up, fingerCount="+fingerCount);
//        if(fingerCount == 1) {
//            tokenUp();
//        } else {
//            dwellStartTime = System.currentTimeMillis();
//        }
//        FINGERS_LOCATION[cursorID] = null;
//        for (int i = 0; i < FINGERS_LOCATION.length; i++) {
//            FINGERS_LOCATION_AT_TIMER_START_TIME[i] = FINGERS_LOCATION[i];
//        }
//        bufferSqueeze.setToInit(true);
//    }

//    (R) I cancel the bending when add/remove fingers but only when there is one finger so if I add finger (2 or more finger) I cancel the timer
    public void fingerUp(MotionEvent event, int index, int id) {
        int cursorID = id;
        int fingerCount = 0;
        // (R) cancel the timer


        for (int i = 0; i < FINGERS_LOCATION.length; i++) {
            if(FINGERS_LOCATION[i] != null) {
                fingerCount++;
            }
        }
//        Log.v("TOUCHTOKENS", "Finger up, fingerCount="+fingerCount);
        if(fingerCount == 1) {
            tokenUp();
        } else {
            dwellStartTime = System.currentTimeMillis();
        }
        FINGERS_LOCATION[cursorID] = null;
        for (int i = 0; i < FINGERS_LOCATION.length; i++) {
            FINGERS_LOCATION_AT_TIMER_START_TIME[i] = FINGERS_LOCATION[i];
        }
        bufferSqueeze.setToInit(true);
        //(R) again cancel the timer
        if(fingerCount != 1 )
        {
            timerBending.cancel();
            Log.v("debugging","cancel timer");
        }
    }

//    public boolean onTouchEvent(MotionEvent event) {
//        int index = MotionEventCompat.getActionIndex(event);
//        int action = MotionEventCompat.getActionMasked(event);
//        int id = MotionEventCompat.getPointerId(event, index);
//        if(debug) {
//            Log.v("TOUCHTOKENS", action+" ("+event.getX(index)+", "+event.getY(index)+") "+id+" - "+index);
//        }
//        switch (action) {
//            case MotionEvent.ACTION_DOWN:
//                fingerDown(event, index, id);
//                break;
//            case MotionEvent.ACTION_POINTER_DOWN:
//                fingerDown(event, index, id);
//                break;
//            case MotionEvent.ACTION_MOVE:
//                //fingerMove(event, index, id);
//                int pointerCount = event.getPointerCount();
//                for(int i = 0; i < pointerCount; ++i) {
//                    index = i;
//                    id = MotionEventCompat.getPointerId(event, index);
//                    if(id < 10){
//                        fingerMove(event, index, id);
//                    }
//                }
//                break;
//            case MotionEvent.ACTION_UP:
//                fingerUp(event, index, id);
//                break;
//            case MotionEvent.ACTION_POINTER_UP:
//                fingerUp(event, index, id);
//                break;
//            default:
//                return true;
//        }
//        if(debug) {
//            String points = "";
//            for (int i = 0; i < FINGERS_LOCATION.length; i++) {
//                if(FINGERS_LOCATION[i] != null) {
//                    points += " "+FINGERS_LOCATION[i];
//                }
//            }
//            Log.v("TOUCHTOKENS", "points="+points);
//        }
//        invalidate();
//        return true;
//    }

//    (R) I just allow the timer to start when There is a token on the surface and to be sure that when there is no token I dont allow the timer to start, sometimes enve when I remove all fingers the timer start and I need to cancel when there is no fingers
    public boolean onTouchEvent(MotionEvent event) {
        int index = MotionEventCompat.getActionIndex(event);
        int action = MotionEventCompat.getActionMasked(event);
        int id = MotionEventCompat.getPointerId(event, index);
        if(debug) {
            Log.v("TOUCHTOKENS", action+" ("+event.getX(index)+", "+event.getY(index)+") "+id+" - "+index);
        }
        switch (action) {
            case MotionEvent.ACTION_DOWN:
                fingerDown(event, index, id);
                break;
            case MotionEvent.ACTION_POINTER_DOWN:
                fingerDown(event, index, id);
                break;
            case MotionEvent.ACTION_MOVE:
                //fingerMove(event, index, id);
                int pointerCount = event.getPointerCount();
                for(int i = 0; i < pointerCount; ++i) {
                    index = i;
                    id = MotionEventCompat.getPointerId(event, index);
                    if(id < 10){
                        fingerMove(event, index, id);
                    }
                }
                break;
            case MotionEvent.ACTION_UP:
                //(R) cancel and dont allow to start the time. I do not understar how it is possible that the timer start when there is no fingers on the surface so I put and boolean to dont start the timer
                timerBending.cancel();
                allowingStartTimer = false;
                Log.v("debugging", "cancel timer in ACTION UP");
                fingerUp(event, index, id);
                break;
            case MotionEvent.ACTION_POINTER_UP:
                fingerUp(event, index, id);
                break;
            default:
                return true;
        }
        if(debug) {
            String points = "";
            for (int i = 0; i < FINGERS_LOCATION.length; i++) {
                if(FINGERS_LOCATION[i] != null) {
                    points += " "+FINGERS_LOCATION[i];
                }
            }
            Log.v("TOUCHTOKENS", "points="+points);
        }
        invalidate();
        return true;
    }
    public double getTokenInitOrientation(TokenTemplate transformedTemplate) {
        return transformedTemplate.getOrientation();
    }

    public double getTokenRelOrientation(TokenTemplate transformedTemplate, ArrayList<TouchPoint> notchPoints) {
        TouchPoint transformedTemplateCenter = transformedTemplate.getCenter();
        TouchPoint transformedTemplate1 = transformedTemplate.getPoints().get(0);
        TouchPoint vTransformedTemplate = new TouchPoint(
                transformedTemplate1.x - transformedTemplateCenter.x,
                transformedTemplate1.y - transformedTemplateCenter.y);
        TouchPoint currentCenter = getTokenCenter(transformedTemplate, notchPoints);
        currentCenter = new TouchPoint(currentCenter.x*viewWidthMm, currentCenter.y*viewHeightMm);
        TouchPoint current1 = notchPoints.get(0);
        current1 = new TouchPoint(current1.x*viewWidthMm, current1.y*viewHeightMm);
        TouchPoint vCurrent = new TouchPoint(
                current1.x - currentCenter.x,
                current1.y - currentCenter.y);
        double angle = TouchTokenRecognizer.angleBetweenVectors(vCurrent, vTransformedTemplate);
        return angle;
    }

    public TouchPoint getTokenCenter(TokenTemplate transformedTemplate, ArrayList<TouchPoint> notchPoints) {
        TouchPoint centerTemplate = new TouchPoint(transformedTemplate.getCenter().x, transformedTemplate.getCenter().y);
        TouchPoint pt1Template = new TouchPoint(transformedTemplate.getPoints().get(0).x, transformedTemplate.getPoints().get(0).y);
        TouchPoint pt2Template = new TouchPoint(transformedTemplate.getPoints().get(1).x, transformedTemplate.getPoints().get(1).y);
        TouchPoint v1CenterTemplate = new TouchPoint(centerTemplate.x - pt1Template.x, centerTemplate.y - pt1Template.y);
        TouchPoint v12Template = new TouchPoint(pt2Template.x - pt1Template.x, pt2Template.y - pt1Template.y);

        double angle21Center = TouchTokenRecognizer.angleBetweenVectors(v12Template, v1CenterTemplate);

        TouchPoint pt1Token = notchPoints.get(0);
        pt1Token = new TouchPoint(pt1Token.x*viewWidthMm, pt1Token.y*viewHeightMm);
        TouchPoint pt2Token = notchPoints.get(1);
        pt2Token = new TouchPoint(pt2Token.x*viewWidthMm, pt2Token.y*viewHeightMm);
        TouchPoint v12Token = new TouchPoint(pt2Token.x - pt1Token.x, pt2Token.y - pt1Token.y);
        double x12Token = TouchTokenRecognizer.angleBetweenVectors(new TouchPoint(1, 0), v12Token);

        double length1Center =
                Math.sqrt((centerTemplate.x - pt1Template.x) * (centerTemplate.x - pt1Template.x) +
                        (centerTemplate.y - pt1Template.y) * (centerTemplate.y - pt1Template.y));

        TouchPoint centerToken = new TouchPoint(length1Center, 0);
        centerToken.setLocation(centerToken.x + pt1Token.getX(), centerToken.y + pt1Token.getY());
        double angle = x12Token + angle21Center;
        TouchTokenRecognizer.rotateBy(centerToken, pt1Token, angle, centerToken);
        centerToken.setLocation(centerToken.x/viewWidthMm, centerToken.y/viewHeightMm);
        return centerToken;
    }

    private double sumDistances(ArrayList<TouchPoint> fingers, TouchPoint center) {
        double d = 0;
        for (TouchPoint p : fingers) {
            if(p != null) {
                d = d + center.distance(p);
            }
        }
        return d;
    }

    protected void initTokenState() {

        ArrayList<TouchPoint> inputInMms = new ArrayList<TouchPoint>(); // the recognizer stores templates in mm
        for (int i = 0; i < FINGERS_LOCATION.length; i++) {
            if(FINGERS_LOCATION[i] != null) {
                inputInMms.add(new TouchPoint(FINGERS_LOCATION[i].x * viewWidthMm, FINGERS_LOCATION[i].y * viewHeightMm));
            }
        }

        currentTemplate = MainActivity.RECOGNIZER.sortTemplatePoints(inputInMms, MainActivity.RECOGNIZER.getTemplate(currentTemplate.getTokenID()));

        notchPointsAtDown = new ArrayList<TouchPoint>();
        for (TouchPoint touchPoint : currentTemplate.getPoints()) {
            notchPointsAtDown.add(new TouchPoint(touchPoint.x / viewWidthMm, touchPoint.y / viewHeightMm));
        }

        // init buffers
        TouchPoint tokenCentroidMm = TouchTokenRecognizer.centroid(inputInMms);
        int fingerCount = inputInMms.size();
        double distance = sumDistances(inputInMms, tokenCentroidMm) / fingerCount;
        bufferBend.bufferClear();
        bufferSqueeze.init(distance, fingerCount);
        bufferSurface.bufferClear();
    }

    /*
    protected void processTokenEvent(String tokenID, TokenTemplate transformedTemplate, ArrayList<TouchPoint> points, double deviceWidth, double deviceHeight) {

        // points are touch points in Normalized Device Space (NDS)

        synchronized(points) {

            double meanDistanceToCentroid = 0;

            ArrayList<TouchPoint> touchPointsNDS = new ArrayList<TouchPoint>();
            int pointsCount = 0;
            for (Iterator<TouchPoint> iterator = points.iterator(); iterator.hasNext();) {
                TouchPoint p = iterator.next();
                touchPointsNDS.add(p);
                if(p != null) {
                    pointsCount++;
                }
            }

            if(pointsCount >= 1 && pointsCount <= 3 ) {
                ArrayList<TouchPoint> touchPointsMm = new ArrayList<TouchPoint>();
                for(int i = 0; i < touchPointsNDS.size(); i++) {
                    TouchPoint p = touchPointsNDS.get(i);
                    if (p != null) {
                        touchPointsMm.add(new TouchPoint(p.getX()*deviceWidth,p.getY()*deviceHeight));
                    }
                }
                TouchPoint tokenCentroidMm = null;
                if (pointsCount > 1) {
                    tokenCentroidMm = TouchTokenRecognizer.centroid(touchPointsMm);
                    meanDistanceToCentroid = sumDistances(touchPointsMm,tokenCentroidMm) / pointsCount;
                }

                long currentTime = System.currentTimeMillis();

                if(pointsCount == 2 || pointsCount == 3) {
                    bufferSurface.updateBuffer(currentTime, meanDistanceToCentroid, pointsCount);
                    bufferBend.updateBuffer(currentTime, meanDistanceToCentroid, pointsCount);
                }

                timeLastEvent = currentTime;

                if(pointsCount == 1) {
                    insufficientInfo = true;
                    if(!bending) {

                        bending = bufferBend.bend();
                        if(bending) {
                            onTokenBent(
                                    tokenID,
                                    points,
                                    lastReliableTokenPoints,
                                    deviceWidth,
                                    deviceHeight,
                                    transformedTemplate,
                                    insufficientInfo);
                        }
                    }
                    lastPointCount = 1;
                    return;
                }

                if(pointsCount == 3 && insufficientInfo) {
                    // we've been through a state where the number of contact points was less than three, it is safer to reinitialize the token state based on current touch points
                    if(pointsCount == 3) {
                        initTokenState();
                    }
                    insufficientInfo = false;
                }

                if(bending) {
                    bending = false;
                    onTokenUnbent(
                            tokenID,
                            points,
                            lastReliableTokenPoints,
                            deviceWidth,
                            deviceHeight,
                            transformedTemplate,
                            insufficientInfo);
                    lastPointCount = pointsCount;
                    return;
                }

                lastPointCount = pointsCount;

                if(pointsCount > 1) {

                    if(bufferSqueeze.isToInit()) {
                        bufferSqueeze.init(meanDistanceToCentroid, pointsCount);
                    }

                    bufferSqueeze.updateBuffer(currentTime, meanDistanceToCentroid);
                    if(bufferSqueeze.getState() == SqueezeState.INITSQUEEZE
                            || bufferSqueeze.getState() == SqueezeState.ENDSQUEEZE) {
                        if(pointsCount == 2) {
                            complementNotchPoints(notchPointsAtDown, touchPointsNDS, deviceWidth, deviceHeight);
                        }
                        if(bufferSqueeze.getState() == SqueezeState.INITSQUEEZE) {
                            onTokenPressed(
                                    tokenID,
                                    points,
                                    touchPointsNDS,
                                    deviceWidth,
                                    deviceHeight,
                                    transformedTemplate,
                                    insufficientInfo);
                        } else {
                            onTokenReleased(
                                    tokenID,
                                    points,
                                    touchPointsNDS,
                                    deviceWidth,
                                    deviceHeight,
                                    transformedTemplate,
                                    insufficientInfo);
                        }
                        return;
                    }
                }

                if(pointsCount == 2 && !insufficientInfo) {
                    complementNotchPoints(notchPointsAtDown, touchPointsNDS, deviceWidth, deviceHeight);
                    pointsCount = 3;
                }

                if(pointsCount == 3) {
                    if(insufficientInfo) {
                        onTokenMoved(
                                tokenID,
                                points,
                                lastReliableTokenPoints,
                                deviceWidth,
                                deviceHeight,
                                transformedTemplate,
                                insufficientInfo);
                    } else {
                        onTokenMoved(
                                tokenID,
                                points,
                                touchPointsNDS,
                                deviceWidth,
                                deviceHeight,
                                transformedTemplate,
                                insufficientInfo);
                    }

                    lastReliableTokenPoints = new ArrayList<TouchPoint>(touchPointsNDS);
                }
            }
        }
    }
*/

    /*
        (R) In this method I added a timer for avoiding problem when you remove all fingers
        And I change the comparator when the user squeeze the token, now only it works when number of contact points are 3
        the part of the code I change I comented before with (R)
    */
    protected void processTokenEvent(String tokenID, TokenTemplate transformedTemplate, ArrayList<TouchPoint> points, double deviceWidth, double deviceHeight) {

        // points are touch points in Normalized Device Space (NDS)

        synchronized(points) {

            currentPoint = points;


            double meanDistanceToCentroid = 0;

            ArrayList<TouchPoint> touchPointsNDS = new ArrayList<TouchPoint>();
            int pointsCount = 0;
            for (Iterator<TouchPoint> iterator = points.iterator(); iterator.hasNext();) {
                TouchPoint p = iterator.next();
                touchPointsNDS.add(p);
                if(p != null) {
                    pointsCount++;
                }
            }

            if(pointsCount >= 1 && pointsCount <= 3 ) {
                ArrayList<TouchPoint> touchPointsMm = new ArrayList<TouchPoint>();
                for(int i = 0; i < touchPointsNDS.size(); i++) {
                    TouchPoint p = touchPointsNDS.get(i);
                    if (p != null) {
                        touchPointsMm.add(new TouchPoint(p.getX()*deviceWidth,p.getY()*deviceHeight));
                    }
                }
                TouchPoint tokenCentroidMm = null;
                if (pointsCount > 1) {
                    tokenCentroidMm = TouchTokenRecognizer.centroid(touchPointsMm);
                    meanDistanceToCentroid = sumDistances(touchPointsMm,tokenCentroidMm) / pointsCount;
                }

                long currentTime = System.currentTimeMillis();

                if(pointsCount == 2 || pointsCount == 3) {
                    bufferSurface.updateBuffer(currentTime, meanDistanceToCentroid, pointsCount);
                    bufferBend.updateBuffer(currentTime, meanDistanceToCentroid, pointsCount);
                }

                timeLastEvent = currentTime;

                if(pointsCount == 1) {
                    insufficientInfo = true;
                    if(!bending & allowingStartTimer ) {
                        //(R) start the timer
                        Log.v("debugging","start timer");
                        timerBending.start();
                        allowingStartTimer = false;
                        //(R) I will activate onTokenBen in the timer in the timer
//                        bending = bufferBend.bend();
//                        if(bending) {
//                            onTokenBent(
//                                    tokenID,
//                                    points,
//                                    lastReliableTokenPoints,
//                                    deviceWidth,
//                                    deviceHeight,
//                                    transformedTemplate,
//                                    insufficientInfo);
//                        }
                    }
                    lastPointCount = 1;
                    return;
                }

                if(pointsCount == 3 && insufficientInfo) {
                    // we've been through a state where the number of contact points was less than three, it is safer to reinitialize the token state based on current touch points
                    if(pointsCount == 3) {
                        initTokenState();
                    }
                    insufficientInfo = false;
                }

                if(bending) {
                    bending = false;
                    onTokenUnbent(
                            tokenID,
                            points,
                            lastReliableTokenPoints,
                            deviceWidth,
                            deviceHeight,
                            transformedTemplate,
                            insufficientInfo);
                    lastPointCount = pointsCount;
//                    (R) allow again to start the timer run
                    allowingStartTimer = true;
                    return;
                }
                lastPointCount = pointsCount;

                //(R) only we activate the sueeze when number of contact points are more than 2
                if(pointsCount > 2) {

                    if(bufferSqueeze.isToInit()) {
                        bufferSqueeze.init(meanDistanceToCentroid, pointsCount);
                    }
                    bufferSqueeze.updateBuffer(currentTime, meanDistanceToCentroid);
                    if(bufferSqueeze.getState() == SqueezeState.INITSQUEEZE
                            || bufferSqueeze.getState() == SqueezeState.ENDSQUEEZE) {
                        // (R) this is not necesary
                        if(pointsCount == 2) {
                            complementNotchPoints(notchPointsAtDown, touchPointsNDS, deviceWidth, deviceHeight);
                        }
                        if(bufferSqueeze.getState() == SqueezeState.INITSQUEEZE) {
                            onTokenPressed(
                                    tokenID,
                                    points,
                                    touchPointsNDS,
                                    deviceWidth,
                                    deviceHeight,
                                    transformedTemplate,
                                    insufficientInfo);
                        } else {
                            onTokenReleased(
                                    tokenID,
                                    points,
                                    touchPointsNDS,
                                    deviceWidth,
                                    deviceHeight,
                                    transformedTemplate,
                                    insufficientInfo);
                        }
                        return;
                    }
                }

                if(pointsCount == 2 && !insufficientInfo) {
                    complementNotchPoints(notchPointsAtDown, touchPointsNDS, deviceWidth, deviceHeight);
                    pointsCount = 3;
                }

                if(pointsCount == 3) {
                    if(insufficientInfo) {
                        onTokenMoved(
                                tokenID,
                                points,
                                lastReliableTokenPoints,
                                deviceWidth,
                                deviceHeight,
                                transformedTemplate,
                                insufficientInfo);
                    } else {
                        onTokenMoved(
                                tokenID,
                                points,
                                touchPointsNDS,
                                deviceWidth,
                                deviceHeight,
                                transformedTemplate,
                                insufficientInfo);
                    }

                    lastReliableTokenPoints = new ArrayList<TouchPoint>(touchPointsNDS);
                }
            }
        }
    }

}
